from datetime import datetime
from enum import StrEnum, auto
from pathlib import Path
from typing import Any, Protocol, Type
from zoneinfo import ZoneInfo

from openai import OpenAI
from openai.types.chat import ChatCompletion
from openai.types.completion_usage import CompletionUsage

from llm_chat.models import Conversation, Message, Role
from llm_chat.settings import Model, OpenAISettings

INITIAL_SYSTEM_MESSAGES = [
    Message(
        role=Role.SYSTEM,
        content=(
            "Always respond in Markdown formatting, using code blocks with the"
            "appropriate language tags."
        ),
    ),
]


def save_conversation(
    conversation: Conversation, history_dir: Path, dt: datetime
) -> None:
    """Store a conversation in the history directory."""
    if conversation.prompt_tokens == 0:
        return

    dt_str = dt.strftime("%Y%m%d%H%M%S")
    name_str = f"-{conversation.name}" if conversation.name else ""
    filename = f"{dt_str}{name_str}"

    history_dir.mkdir(parents=True, exist_ok=True)

    path = history_dir / f"{filename}.json"
    with path.open(mode="w") as f:
        f.write(conversation.model_dump_json(indent=2))


class Token(StrEnum):
    """Token type for the OpenAI chat."""

    COMPLETION = auto()
    PROMPT = auto()


class ChatProtocol(Protocol):
    """Protocol for chat classes."""

    conversation: Conversation

    @property
    def cost(self) -> float:
        """Get the cost of the conversation."""

    @property
    def name(self) -> str:
        """Get the name of the conversation."""

    @property
    def settings(self) -> OpenAISettings:
        """Get OpenAI chat settings."""

    def save(self) -> None:
        """Save the conversation to the history directory."""

    def send_message(self, prompt: str) -> str:
        """Send a message to the assistant."""


class Chat:
    """Interface class for OpenAI's ChatGPT chat API.

    Arguments:
        settings (optional): Settings for the chat. Defaults to reading from
            environment variables.
        context (optional): Context for the chat. Defaults to an empty list.
    """

    _pricing: dict[Model, dict[Token, float]] = {
        Model.GPT3: {
            Token.COMPLETION: 0.002,
            Token.PROMPT: 0.0015,
        },
        Model.GPT4: {
            Token.COMPLETION: 0.06,
            Token.PROMPT: 0.03,
        },
    }

    def __init__(
        self,
        settings: OpenAISettings | None = None,
        context: list[Message] = [],
        name: str = "",
        initial_system_messages: bool = True,
    ) -> None:
        self._settings = settings
        self.conversation = Conversation(
            messages=INITIAL_SYSTEM_MESSAGES + context
            if initial_system_messages
            else context,
            model=self.settings.model,
            temperature=self.settings.temperature,
            name=name,
        )
        self._start_time = datetime.now(tz=ZoneInfo("UTC"))
        self._client = OpenAI(
            api_key=self.settings.api_key,
        )

    @classmethod
    def load(
        cls, path: Path, api_key: str | None = None, history_dir: Path | None = None
    ) -> ChatProtocol:
        """Load a chat from a file."""
        with path.open() as f:
            conversation = Conversation.model_validate_json(f.read())
        args: dict[str, Any] = {
            "model": conversation.model,
            "temperature": conversation.temperature,
        }
        if api_key is not None:
            args["api_key"] = api_key
        if history_dir is not None:
            args["history_dir"] = history_dir

        settings = OpenAISettings(**args)
        return cls(
            settings=settings,
            context=conversation.messages,
            initial_system_messages=False,
        )

    @property
    def settings(self) -> OpenAISettings:
        """Get OpenAI chat settings."""
        if self._settings is None:
            self._settings = OpenAISettings()
        return self._settings

    @property
    def cost(self) -> float:
        """Get the cost of the conversation."""
        return self.conversation.cost

    @property
    def name(self) -> str:
        """Get the name of the conversation."""
        return self.conversation.name

    def _make_request(self, message: str) -> ChatCompletion:
        """Send a request to the OpenAI API.

        TODO: Add error handling.
        TODO: Pop message from messages if error occurs.
        """
        self.conversation.messages.append(Message(role=Role.USER, content=message))

        response: ChatCompletion = self._client.chat.completions.create(
            model=self.settings.model,
            temperature=self.settings.temperature,
            messages=self.conversation.model_dump()["messages"],
            stream=False,
        )

        return response

    def _calculate_cost(self, usage: CompletionUsage) -> None:
        """Calculate the cost of a request."""
        self.conversation.completion_tokens += usage.completion_tokens
        self.conversation.prompt_tokens += usage.prompt_tokens
        self.conversation.cost = round(
            (self.conversation.completion_tokens / 1000)
            * self._pricing[self.settings.model][Token.COMPLETION]
            + (self.conversation.prompt_tokens / 1000)
            * self._pricing[self.settings.model][Token.PROMPT],
            6,
        )

    def save(self) -> None:
        """Save the conversation to the history directory."""
        save_conversation(
            conversation=self.conversation,
            history_dir=self.settings.history_dir,
            dt=self._start_time,
        )

    def send_message(self, prompt: str) -> str:
        """Send a message to the assistant.

        TODO: Add error handling.
        """
        request_response = self._make_request(prompt)

        if request_response.usage is not None:
            self._calculate_cost(request_response.usage)
        response = request_response.choices[0].message.content
        self.conversation.messages.append(
            Message(role=Role.ASSISTANT, content=response if response else "")
        )
        return response if response else ""


def get_chat(
    settings: OpenAISettings | None = None, context: list[Message] = [], name: str = ""
) -> ChatProtocol:
    """Get a chat object."""
    return Chat(settings=settings, context=context, name=name)


def get_chat_class() -> Type[Chat]:
    """Get the chat class."""
    return Chat
